// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#include <fstream>
#include <chrono>
#include <filesystem>
#include <vector>
#include <algorithm>
#include "sample_npu_utils.h"
#include "EsHwPreProcess.h"

#define MAX_WIDTH 3840
#define MAX_HEIGHT 2160

std::atomic<bool> EsHwPreProcess::mInit{false};

EsHwPreProcess::EsHwPreProcess(ES_U16 devId) : mGrpCount(1), mGrpId(0), mChnId(0), mDieId(devId) {}

EsHwPreProcess::~EsHwPreProcess() { release(); }

bool EsHwPreProcess::initializeFromJson(uint32_t devId, const std::string &cfgFilePath)
{
    bool success = parsePreProcessParamsFromJson(cfgFilePath, mPreprocessParams);
    if (!success) return success;

    // set normalization params
    memset(&mNormParams, 0, sizeof(mNormParams));
    mNormParams.normalizationMode = VPS_NORMALIZATION_Z_SCORE;

    ES_FLOAT mean1, mean2, mean3;
    ES_FLOAT std1, std2, std3;
    ES_FLOAT scale;

    if (mPreprocessParams.norm) {
        scale = 255;
    } else {
        scale = 1;
    }

    mean1 = mPreprocessParams.meanValues[0] * scale;
    mean2 = mPreprocessParams.meanValues[1] * scale;
    mean3 = mPreprocessParams.meanValues[2] * scale;
    std1 = 1 / (mPreprocessParams.stdValues[0] * scale);
    std2 = 1 / (mPreprocessParams.stdValues[1] * scale);
    std3 = 1 / (mPreprocessParams.stdValues[2] * scale);
    if (mPreprocessParams.inputFormat == "BGR") {
        mNormParams.meanValue.r = *((ES_U32 *)&mean3);
        mNormParams.meanValue.g = *((ES_U32 *)&mean2);
        mNormParams.meanValue.b = *((ES_U32 *)&mean1);
        mNormParams.stdReciprocal.r = *((ES_U32 *)&std3);
        mNormParams.stdReciprocal.g = *((ES_U32 *)&std2);
        mNormParams.stdReciprocal.b = *((ES_U32 *)&std1);
    } else {
        mNormParams.meanValue.r = *((ES_U32 *)&mean1);
        mNormParams.meanValue.g = *((ES_U32 *)&mean2);
        mNormParams.meanValue.b = *((ES_U32 *)&mean3);
        mNormParams.stdReciprocal.r = *((ES_U32 *)&std1);
        mNormParams.stdReciprocal.g = *((ES_U32 *)&std2);
        mNormParams.stdReciprocal.b = *((ES_U32 *)&std3);
    }

    ES_FLOAT reciprocal = 1 / mPreprocessParams.quantizationScale;
    mNormParams.stepReciprocal = *((ES_U32 *)&reciprocal);
    // mNormParams.bByPassQuantization = ES_TRUE;

    if (!mInit.load()) {
        mInit.store(true);
        ES_S32 i;
        ES_S32 ret = ES_SUCCESS;
        /************************************************
        init VPS
        *************************************************/
        ret = SAMPLE_COMM_VPS_Init();
        if (ES_SUCCESS != ret) {
            DEBUG_ERROR("SAMPLE_COMM_VPS_Init failed with %#x!\n", ret);
            goto END1;
        }
    }

    return success;
END1:
    SAMPLE_COMM_VPS_Deinit();

    return false;
}

int32_t EsHwPreProcess::preprocess(VIDEO_FRAME_INFO_S &frame, PREPROCESSED_DATA &outData)
{
    std::lock_guard<std::mutex> lock(mMtx);

    ES_U64 PTS = 0;
    ES_S32 ret;
    ES_FLOAT scale;
    SIZE_S middSize, dstSize;
    PIXEL_FORMAT_E middFmt, dstFmt;
    RECT_S srcRect, middRect, dstRect;

    auto start_time = std::chrono::high_resolution_clock::now();

    VIDEO_FRAME_INFO_S middFrame;
    memset(&middFrame, 0, sizeof(VIDEO_FRAME_INFO_S));
    middSize.width = mPreprocessParams.resizeShape;
    middSize.height = mPreprocessParams.resizeShape;
    middFmt = PIXEL_FORMAT_R8G8B8;
    ret = SAMPLE_COMM_INIT_BUFFER(&middFrame, middSize, middFmt, mDieId);
    if (ret < 0) {
        SAMPLE_PRT("init dstFrame buffer failed, ret: %d\n", ret);
        ret = ES_FAILURE;
        return ret;
    }

    ES_S32 newHeight;
    ES_S32 newWidth;

    if (mPreprocessParams.keepRatio) {
        // fill whole image with fillColor
        RECT_S fillRect = {0, 0, middFrame.videoFrame.width, middFrame.videoFrame.height};
        ES_U32 fillColor = 0xFF727272;
        ret = ES_VPS_Fill(&middFrame.videoFrame, &fillColor, 1, &fillRect, 1, HW_TYPE_HAE);
        if (ret != ES_SUCCESS) {
            DEBUG_ERROR("ES_VPS_Fill failed, %d\n", ret);
        }
        scale = static_cast<ES_FLOAT>(mPreprocessParams.resizeShape) /
                std::max(frame.videoFrame.width, frame.videoFrame.height);
        if (frame.videoFrame.height > frame.videoFrame.width) {
            newHeight = mPreprocessParams.resizeShape;
            newWidth = static_cast<ES_S32>(frame.videoFrame.width * scale);
        } else {
            newHeight = static_cast<ES_S32>(frame.videoFrame.height * scale);
            newWidth = mPreprocessParams.resizeShape;
        }
    } else {
        newHeight = middFrame.videoFrame.height;
        newWidth = middFrame.videoFrame.width;
    }

    memset(&middRect, 0, sizeof(middRect));
    if (frame.videoFrame.width > frame.videoFrame.height) {
        middRect.x = 0;
        middRect.y = (middSize.height - newHeight) / 2;
        middRect.width = newWidth;
        middRect.height = newHeight;
    } else {
        middRect.x = (middSize.width - newWidth) / 2;
        middRect.y = 0;
        middRect.width = newWidth;
        middRect.height = newHeight;
    }

    VPS_PROPERTY_S prop = {};
    prop.type = VPS_PROPERTY_RESIZE_METHOD;
    // VPS_RESIZE_STRETCHBLIT, VPS_RESIZE_FILTERBLIT, VPS_RESIZE_BILINEAR, VPS_RESIZE_BICUBIC
    prop.u.resizeMethod = VPS_RESIZE_FILTERBLIT;
    ret = ES_VPS_SetProperty(&prop);
    if (ret != ES_SUCCESS) {
        DEBUG_ERROR("Set resize method [%d] failed\n", prop.u.resizeMethod);
    }

    // resize frame(originW, originH) to middle frame
    ret = ES_VPS_CropResize(&frame.videoFrame, &middFrame.videoFrame, ES_NULL, &middRect, 0, HW_TYPE_HAE);
    if (ret != ES_SUCCESS) {
        DEBUG_ERROR("ES_VPS_Resize failed, %d\n", ret);
    }

    // 2D support bgr2grb, not rgb2bgr, so we should adjust mean and std, then set input to bgr.
    if (mPreprocessParams.inputFormat == "BGR") {
        middFrame.videoFrame.pixelFormat = PIXEL_FORMAT_B8G8R8;
    }
    VIDEO_FRAME_INFO_S dstFrame;
    memset(&dstFrame, 0, sizeof(VIDEO_FRAME_INFO_S));
    dstFrame.poolId = ES_VB_INVALID_POOLID;
    dstFrame.modId = ES_ID_USER;
    dstFrame.videoFrame.width = mPreprocessParams.cropWidth;
    dstFrame.videoFrame.height = mPreprocessParams.cropWidth;
    dstFrame.videoFrame.fd = outData.fd;

    if (mPreprocessParams.channelFirst) {
        dstFrame.videoFrame.pixelFormat = PIXEL_FORMAT_R8G8B8I_PLANAR;
        dstFrame.videoFrame.stride[0] = mPreprocessParams.cropWidth;
        dstFrame.videoFrame.stride[1] = mPreprocessParams.cropWidth;
        dstFrame.videoFrame.stride[2] = mPreprocessParams.cropWidth;

        dstFrame.videoFrame.offset[0] = outData.offset;
        dstFrame.videoFrame.offset[1] = outData.offset + (mPreprocessParams.cropWidth * mPreprocessParams.cropWidth);
        dstFrame.videoFrame.offset[2] =
            outData.offset + (mPreprocessParams.cropWidth * mPreprocessParams.cropWidth * 2);
    } else {
        dstFrame.videoFrame.pixelFormat = PIXEL_FORMAT_R8G8B8I;
        dstFrame.videoFrame.stride[0] = mPreprocessParams.cropWidth * 3;
        dstFrame.videoFrame.stride[1] = 0;
        dstFrame.videoFrame.stride[2] = 0;

        dstFrame.videoFrame.offset[0] = outData.offset;
        dstFrame.videoFrame.offset[1] = 0;
        dstFrame.videoFrame.offset[2] = 0;
    }

    memset(&srcRect, 0, sizeof(srcRect));
    srcRect.x = (middFrame.videoFrame.width - dstFrame.videoFrame.width) / 2;
    srcRect.y = (middFrame.videoFrame.height - dstFrame.videoFrame.height) / 2;
    srcRect.width = dstFrame.videoFrame.width;
    srcRect.height = dstFrame.videoFrame.height;

    DEBUG_INFO("src width:%d, height:%d\n", frame.videoFrame.width, frame.videoFrame.height);
    DEBUG_INFO("dst width:%d, height:%d\n", dstFrame.videoFrame.width, dstFrame.videoFrame.height);
    DEBUG_INFO("src rect: w:%d, h:%d, x:%d, y:%d\n", srcRect.width, srcRect.height, srcRect.x, srcRect.y);

    // Normalization
    ret = ES_VPS_Normalization(&middFrame.videoFrame, &dstFrame.videoFrame, &mNormParams, &srcRect, NULL);
    if (ret != ES_SUCCESS) {
        DEBUG_ERROR("ES_VPS_Normalization return with 0x%x!\n", ret);
    }

    ////////////////////////////////////////////////////////////////////////////////////////
    SAMPLE_COMM_DEINIT_BUFFER(&middFrame);

    auto end_time = std::chrono::high_resolution_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::duration<double, std::milli>>(end_time - start_time);

    DEBUG_INFO("EsHwPreProcess::%s Preprocess cost:%.4f\n", __func__, duration.count());

    return ret;
}

void EsHwPreProcess::release()
{
    if (mInit.load()) {
        mInit.store(false);
        DEBUG_INFO("EsHwPreProcess::release\n");
        SAMPLE_COMM_VPS_Deinit();
    }
}
